import json
from ha.interfaces.gpioInterface import *
from ha import *

class ApplianceInterface(Interface):
    def __init__(self, name, interface):
        Interface.__init__(self, name, interface)
        self.stateFile = stateDir+name+".state"

    def start(self):
        # read the state from the file if it exists
        try:
            self.state = json.load(open(self.stateFile))
            debug('debugState', self.name, "reading", self.stateFile, self.state)
        except:
            self.state = {}
        # set the state of all the devices on the interface
        for addr in list(self.state.keys()):
            # addr must be converted to int because json stringified it when dumped
            self.interface.write(int(addr), self.state[addr])

    def read(self, addr):
        try:
            return self.state[str(addr)]    # keys are saved as strings
        except:
            # if the device does not exist, create it and set the state to 0
            self.write(addr, 0)
            return 0

    def write(self, addr, value):
        # set the state of the device and save it
        self.interface.write(addr, value)
        self.state[str(addr)] = value   # keys are saved as strings
        debug('debugState', self.name, "writing", self.stateFile, self.state)
        json.dump(self.state, open(self.stateFile, "w"))




from ha import *

# TMP102 temp sensor

class TMP102Interface(Interface):
    def __init__(self, name, interface):
        Interface.__init__(self, name, interface)

    def read(self, addr):
        debug('debugTemp', self.name, "read", addr)
        try:
            t = self.interface.readWord((addr, 0))
            return (float(((t&0x00ff)<<4) | ((t&0xf000)>>12)) * .0625) * 9 / 5 + 32
        except:
            return 0




# Send RGB light colors over a serial interface
from ha import *

syn = 0x16
def checksum(digits):
    sum = 0
    for digit in digits:
        sum += digit
    return sum % 256

class SerialRGBInterface(Interface):
    def __init__(self, name, interface, **kwargs):
        Interface.__init__(self, name, interface=interface, **kwargs)
        self.bytes = bytearray()

    def write(self, addr, value):
        debug("debugHolidayLights", self.name, "write", value)
        self.bytes.append((value>>16) & 0xff)     # red
        self.bytes.append((value>>8) & 0xff)      # green
        self.bytes.append(value & 0xff)           # blue
        return True

    def show(self):
        msg  = bytearray([syn, syn, syn])                   # sync bytes
        msg += len(self.bytes).to_bytes(1, "little")        # data length
        msg += self.bytes                                   # data
        msg += checksum(self.bytes).to_bytes(1, "little")   # checksum
        debug("debugHolidayLights", self.name, "show", msg)
        self.interface.write(0, msg)
        self.bytes = bytearray()
        return True



from w1thermsensor import W1ThermSensor
from ha import *

# 1wire temp sensor

class DS18B20Interface(Interface):
    def __init__(self, name, interface=None):
        Interface.__init__(self, name, interface)

    def read(self, addr):
        try:
            return float(W1ThermSensor(W1ThermSensor.THERM_SENSOR_DS18B20, addr).get_temperature())
        except Exception as ex:
            log("ds18b20", type(ex).__name__, str(ex))
            return 0



# Calculate wind speed and direction using the Peet Ultimeter Pro Anemometer
timeOut = 10

import time
import threading
from ha import *
from ha.interfaces.gpioInterface import *

# global variables
startTime = time.time() # speed contact closure time
speedTime = 0.0         # delta tSpeed
dirTime = 0.0           # delta tDir

# anemometer contact closure interrupt routine
def speedInterrupt(sensor, state):
    global startTime, speedTime, dirTime
    curTime = time.time()
    if state == 0:
        speedTime = curTime - startTime
        startTime = curTime

# wind vane contact closure interrupt routine
def dirInterrupt(sensor, state):
    global startTime, speedTime, dirTime
    curTime = time.time()
    if state == 0:
        dirTime = curTime - startTime

class WindInterface(Interface):
    def __init__(self, name, interface, anemometer, windVane, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.anemometer = anemometer    # anemometer contact closure sensor
        self.anemometer.interrupt = speedInterrupt
        self.windVane = windVane        # wind vane contact closure sensor
        self.windVane.interrupt = dirInterrupt
        self.speed = 0.0
        self.dir = 0.0
        debug("debugWind", self.name, "anemometer:", self.anemometer.name, "windVane:", self.windVane.name)

    def read(self, addr):
        curTime = time.time() - startTime
        debug("debugWind", self.name, "read", "addr:", addr, "startTime:", startTime, "curTime:", curTime)
        if addr == "speed":
            # if no data for timeOut seconds, there is no wind
            if curTime > timeOut:
                rps = 0
                self.speed = 0.0
            else:
                # calculate revolutions per second
                try:
                    rps = 1.0 / speedTime
                except ZeroDivisionError:
                    rps = 0.0
                # calculate the wind speed
                if rps < .010:
                    self.speed = 0.0
                elif rps < 3.229:   # 0.2 - 8.2 speed
                    self.speed = -0.1095 * rps**2 + 2.9318 * rps - 0.1412
                elif rps < 54.362:  # 8.2 - 136.0 speed
                    self.speed = 0.0052 * rps**2 + 2.1980 * rps + 1.1091
                else:               # 136.0 - 181.5 speed
                    # self.speed = 0.1104 * rps**2 + 9.5685 * rps + 329.87
                    self.speed = 0.0    # throw out values > 136 mph
            debug("debugWind", self.name, "speedTime:", speedTime, "rps:", rps, "speed:", self.speed)
            return int(self.speed + .5)
        elif addr == "dir":
            # calculate direction only if there is wind
            if self.speed > 0.0:
                try:
                    self.dir = (180 + (360 * dirTime / speedTime)) % 360
                except ZeroDivisionError:
                    self.dir = 0.0
                debug("debugWind", self.name, "dirTime:", dirTime, "speedTime:", speedTime, "direction:", self.dir)
            return int(self.dir + .5)
        else:
            return 0.0




import time
import serial
import struct
import threading
from ha import *
from functools import reduce

# ASCII constants
NUL = b'\x00'
DLE = b'\x10'
STX = b'\x02'
ETX = b'\x03'
SOM = b'\xa5'    # start of message
FIL = b'\xff'    # filler

# addresses
ctrlAddr  = b'\x11'

# commands
writeCmd   = b'\x01'
setCtrlCmd = b'\x04'
setModCmd  = b'\x05'
setRunCmd  = b'\x06'
sendCmd    = b'\x07'

# pump constants
pa = 12.50
pb = 1377.22
pc = 75.28

########################################################################################################
# state of the pool pump
########################################################################################################
class PentairInterface(Interface):
    # constructor
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        # in theory there could be multiple pumps on an interface, but assume just one for now
        self.addr = pentairAddr

        # The pump state consists of 4 values:
        # 0 - predefined pump speed 0..3
        # 1 - speed in rpm
        # 2 - power in watts
        # 3 - flow in gpm

        self.state = [0, 0, 0, 0]
        self.manSpeed = 0           # the speed to which the pump has been manually set
        self.reqSpeed = -1          # the speed to which it is being requested to be set manually

    def start(self):
        debug('debugPentairData', self.name, "starting pentair interface", self.interface.name)
        self.interface.start()
        msgThread = PentairMsgThread("pentair msg", self)
        msgThread.start()

    def read(self, addr):
        return self.state[addr]

    def write(self, addr, value):
        if value in range(0, 5):
            self.reqSpeed = value

    def readState(self):
        self.sendMsg(self.addr, setCtrlCmd, b'\xff')
        (dest, cmd, reply) = self.readMsg()
        self.sendMsg(self.addr, sendCmd, b'')
        (src, cmd, stat) = self.readMsg()
        status = struct.unpack("!BBBHHBBBBBBBB", stat)
        self.sendMsg(self.addr, setCtrlCmd, b'\x00')
        (src, cmd, reply) = self.readMsg()
        self.state[1] = status[4]
        self.state[2] = status[3]
        try:
            rpmRatio = status[4]/3450.0
            self.state[3] = (status[3]-pb*rpmRatio**3-pc)/(pa*rpmRatio**2)
        except:
            self.state[3] = 0.0
        # set the speed to the index of the table entry that is next highest to the rpm value
        self.state[0] = 0
        for speed in range(len(pentairSpeeds)):
            if self.state[1] <= pentairSpeeds[speed]:
                self.state[0] = speed
                break

    def setSpeed(self, theSpeed):
        dat = struct.pack("!L",0x03210000+(theSpeed<<3))
        self.sendMsg(self.addr, setCtrlCmd, b'\xff')
        (dest, cmd, reply) = self.readMsg()
        self.sendMsg(self.addr, writeCmd, dat)
        (dest, cmd, reply) = self.readMsg()
#            self.sendMsg(pumpAddr, sendCmd, b'')
#            (src, cmd, stat) = self.readMsg()
#            status = struct.unpack("!BBBHHBBBBBBBB", stat)
        self.sendMsg(self.addr, setCtrlCmd, b'\x00')
        (src, cmd, reply) = self.readMsg()

    def readMsg(self):
        """ Read the next valid message from the serial port.
        Parses and returns the source address, command, and data as a tuple."""
        while True:
            pmsg = PentairMsg(iface=self.interface)
            # stop reading if a message with a valid checksum is read
            if checksum16(pmsg.msg[0:-2]) == pmsg.sum:
                debug('debugPentairData', self.name, "-->", pmsg.printState())
                return (pmsg.src, pmsg.cmd, pmsg.dat)
            else:
                debug('debugPentairData', self.name, "-->", pmsg.printState(),
                                  "*** bad checksum ***")
                return (pmsg.src, pmsg.cmd, pmsg.dat)

    def sendMsg(self, dst, cmd, dat):
        """ Send a message.
        """
        pmsg = PentairMsg(dst=dst, src=ctrlAddr, cmd=cmd, dat=dat)
        debug('debugPentairData', self.name, "<--", pmsg.printState())
        self.interface.write(None, FIL+NUL+FIL+pmsg.msg)

class PentairMsg(object):
    def __init__(self, iface=None, dst=b"", src=b"", cmd=b"", dat=b""):
        if iface != None:    # read the message from the interface and parse it
            self.msg = b""
            while (self.msg != SOM):            # read until \xa5
                self.msg = iface.read(None, 1)
            self.msg += iface.read(None, 5)            # read header
            dataLen = self.msg[-1] # struct.unpack("!B", self.msg[-1])[0]
            self.msg += iface.read(None, dataLen+2)    # read data and checksum
            self.som = self.msg[0:1]
            self.sub = self.msg[1:2]
            self.dst = self.msg[2:3]
            self.src = self.msg[3:4]
            self.cmd = self.msg[4:5]
            self.len = self.msg[5:6]
            self.dat = self.msg[6:-2]
            self.sum = self.msg[-2:]
        else:           # construct the message from arguments
            self.som = SOM
            self.sub = NUL
            self.dst = dst
            self.src = src
            self.cmd = cmd
            self.len = bytes([len(dat)]) # struct.pack("!B", len(dat))
            self.dat = dat
            self.msg = self.som+self.sub+self.dst+self.src+cmd+self.len+self.dat
            self.sum = checksum16(self.msg)
            self.msg = self.msg+self.sum

    def printState(self):
        return self.som.hex()+self.sub.hex()+" "+\
                             self.dst.hex()+" "+self.src.hex()+" "+\
                             self.cmd.hex()+" "+self.len.hex()+" "+\
                             self.dat.hex()+" "+\
                             self.sum.hex()

class PentairMsgThread(LogThread):
    """ Message handling thread.

    """
    def __init__(self, name, interface=None):
        """ Initialize the thread."""
        LogThread.__init__(self, target=self.doMsg)
        self.name = name
        self.interface = interface

    def doMsg(self):
        """ Message handling loop.
        """
        debug('debugThread', self.name, "started")
        debug('debugPentairThread', self.name, "reading status")
        self.interface.readState()
        loopCount = 0
        refreshInterval = 25
        # wake up every second
        while True:
            # set the speed if there was a request to change it
            if self.interface.reqSpeed >= 0:
                debug('debugPentairThread', self.name, "setting speed to", self.interface.reqSpeed)
                self.interface.setSpeed(self.interface.reqSpeed)
                self.interface.manSpeed = self.interface.reqSpeed
                self.interface.reqSpeed = -1
                loopCount = refreshInterval
            # check the status every refresh interval
            if loopCount == refreshInterval:
                # if the pump has been manually set, resend the message to maintain the speed
                if self.interface.manSpeed > 0:
                    debug('debugPentairThread', self.name, "maintaining speed", self.interface.manSpeed)
                    self.interface.setSpeed(self.interface.manSpeed)
#                debug('debugPentairThread', self.name, "reading status")
                self.interface.readState()
                loopCount = 0
            loopCount += 1
            time.sleep(1)
        debug('debugThread', self.name, "terminated")

def checksum16(msg):
    """ Compute the 16 bit checksum of a string of bytes."""
    return struct.pack("!H", reduce(lambda x,y:x+y, list(msg)) % 16384)




import smbus
from ha import *

class I2CInterface(Interface):
    def __init__(self, name, interface=None, event=None, bus=0):
        Interface.__init__(self, name, interface=interface, event=event)
        self.bus = smbus.SMBus(bus)

    def read(self, addr):
        try:
            debug('debugI2C', self.name, "readByte", addr)
            return self.bus.read_byte_data(*addr)
        except OSError:
            return 0

    def readWord(self, addr):
        try:
            debug('debugI2C', self.name, "readWord", addr)
            return self.bus.read_word_data(*addr)
        except OSError:
            return 0

    def readBlock(self, addr, reg, length):
        try:
            debug('debugI2C', self.name, "readBlock", addr, reg, length)
            return self.bus.read_i2c_block_data(addr, reg, length)
        except OSError:
            return [0, 0]

    def write(self, addr, value):
        try:
            debug('debugI2C', self.name, "writeByte", addr, value)
            self.bus.write_byte_data(*addr+(value,))
        except OSError:
            return 0

    def writeWord(self, addr, value):
        try:
            debug('debugI2C', self.name, "writeWord", addr, value)
            self.bus.write_word_data(*addr+(value,))
        except OSError:
            return 0

    def writeBlock(self, addr, reg, list):
        try:
            debug('debugI2C', self.name, "writeBlock", addr, reg, list)
            self.bus.write_i2c_block_data(addr, reg, list)
        except OSError:
            return []

    def writeQuick(self, addr):
        try:
            debug('debugI2C', self.name, "writeQuick", addr)
            self.bus.write_quick(addr)
        except OSError:
            return 0



import json
import requests
import urllib.parse
import sys
from ha import *
from ha.rest.restConfig import *

class RestInterface(Interface):
    def __init__(self, name, interface=None, event=None, serviceAddr="", cache=True, writeThrough=True):
        Interface.__init__(self, name, interface=interface, event=event)
        self.serviceAddr = serviceAddr      # address of the REST service to target (ipAddr:port)
        self.cache = cache                  # cache the states
        self.writeThrough = writeThrough    # cache is write through
        self.enabled = False
        debug('debugRest', self.name, "created", self.serviceAddr)

    def start(self):
        debug('debugRest', self.name, "starting")
        self.enabled = True

    def stop(self):
        if self.enabled:
            self.enabled = False
            debug('debugRest', self.name, "stopping")
            # invalidate the state cache
            for state in list(self.states.keys()):
                self.states[state] = None

    # disable the RestService that uses this interface
    def disableService(self):
        debug('debugRestProxyDisable', self.name, "disabled")
        for sensor in list(self.sensors.values()):
            if sensor.type == "service":    # just disable the service, it will disable the rest
                sensor.disable("REST I/O error")
                break

    # return the state value for the specified sensor address
    # addr is the REST path to the specified resource
    def read(self, addr):
        debug('debugRestStates', self.name, "read", addr, self.states)
        if not self.enabled:
            return None
        # return the state from the cache if it is there, otherwise read it from the service
        if (not self.cache) or (self.states[addr] == None):
            self.states[addr] = self.readRest("/resources/"+addr+"/state")["state"]
        return self.states[addr]

    # get state values of all sensors on this interface
    def getStates(self, path="/states"):
        debug('debugRestStates', self.name, "getStates", "path", path)
        try:
            states = self.readRest(path)[path.split("/")[-1]]
        except KeyError:
            states = {}
        debug('debugRestStates', self.name, "getStates", "states", states)
        self.setStates(states)
        return states

    # set state values of all sensors into the cache
    def setStates(self, states):
        for sensor in list(states.keys()):
            self.states[sensor] = states[sensor]

    # read the json data from the specified path and return a dictionary
    def readRest(self, path):
        debug('debugRestStates', self.name, "readRest", path)
        try:
            url = "http://"+self.serviceAddr+urllib.parse.quote(path)
            debug('debugRestGet', self.name, "GET", url)
            response = requests.get(url, timeout=restTimeout)
            debug('debugRestGet', self.name, "status", response.status_code)
            if response.status_code == 200:
                debug('debugRestGet', self.name, "response", response.json())
                return response.json()
            else:
                log(self.name, "read state status", response.status_code)
                return {}
        except requests.exceptions.Timeout:
            log(self.name, "read state timeout", path)
            self.disableService()
        except Exception as ex:
            logException(self.name+" read exception", ex)
            self.disableService()

    # write the control state to the specified address
    # addr is the REST path to the specified resource
    def write(self, addr, value):
        debug('debugRestStates', self.name, "write", addr, value)
        if self.enabled:
            if self.cache:
                if self.writeThrough:
                    # update the cache
                    self.states[addr] = value
                    self.notify()
                else:
                    # invalidate the cache
                    self.states[addr] = None
            # create a jsonized dictionary
            data=json.dumps({"state": value})
            # data=json.dumps({addr.split("/")[-1]: value})
            return self.writeRest("/resources/"+addr+"/state", data)
        else:
            return False

    # write json data to the specified path
    def writeRest(self, path, data):
        debug('debugRestStates', self.name, "writeRest", path, data)
        try:
            url = "http://"+self.serviceAddr+urllib.parse.quote(path)
            debug('debugRestPut', self.name, "PUT", url, "data:", data)
            response = requests.put(url,
                             headers={"Content-type":"application/json"},
                             data=data)
            debug('debugRestPut', self.name, "status", response.status_code)
            if response.status_code == 200:
                return True
            else:
                log(self.name, "write state status", response.status_code)
                return False
        except Exception as ex:
            logException(self.name+" write exception", ex)
            self.disableService()



audioSink = "alsa_output.0.analog-stereo"
audioSink = "0"

import os
import subprocess
from ha import *

class AudioInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)

    def read(self, addr):
        debug('debugAudio', self.name, "read", addr)
        try:
            if addr == "volume":
                value = subprocess.check_output("ssh pi@localhost pacmd list-sinks|grep volume", shell=True)
                return int(value.split("/")[1].strip(" ").strip("%"))
            elif addr == "mute":
                value = subprocess.check_output("ssh pi@localhost pacmd list-sinks|grep muted", shell=True)
                return 1 if value.split(":")[1].strip(" ").strip("\n") == "yes" else 0
            elif addr == "wifi":
                value = subprocess.check_output("ifconfig wlan0|grep inet\ addr", shell=True)
                return 0 if value == "" else 1
        except:
            return 0

    def write(self, addr, value):
        debug('debugAudio', self.name, "write", addr, value)
        if addr == "volume":
            os.system("ssh pi@localhost pactl set-sink-volume %s %d%%" % (audioSink, value))
        elif addr == "mute":
            os.system("ssh pi@localhost pactl set-sink-mute %s %d" % (audioSink, value))
        elif addr == "wifi":
            if value == 0:
                os.system("/root/carputer/wifi.sh off &")
            else:
                os.system("/root/carputer/wifi.sh on &")





from ha import *

# TC74 temp sensor

class TC74Interface(Interface):
    def __init__(self, name, interface):
        Interface.__init__(self, name, interface)

    def read(self, addr):
        debug('debugTemp', self.name, "read", addr)
        try:
            value = self.interface.read((addr, 0))
            if value > 127:
                value = (256-value) * (-1)
            return float(value) * 9 / 5 + 32
        except:
            return 0



# Global variables that define the environment

import os
import socket
import platform

# determine the hardware this is running on
machine = platform.machine()
if machine in ["armv6l", "armv6l"]:
    platformType = "rpi"
elif machine in ["esp8266", "esp32"]:
    platformType = "esp"
elif machine in ["x86_64"]:
    platformType = "mac"
else:
    platformType = "unknown"

# directory structure
hostname = socket.gethostname()
rootDir = os.path.expanduser("~")+"/"
configDir = rootDir+"conf/"
keyDir = rootDir+"keys/"
stateDir = rootDir+"state/"
soundDir = rootDir+"sounds/"
dataLogDir = "data/"
dataLogFileName = ""

# Localization - define these in the config file
    # latLong = (0.0, 0.0)
    # elevation = 0 # elevation in feet
    # tempScale = "F"

# global variables that must be set here
sysLogging = True




import RPi.GPIO as gpio
from ha import *

# available GPIO pins
if int(gpio.RPI_INFO['REVISION'], 16) < 0x0010:
    bcmPins = [4, 17, 18, 22, 23, 24, 25, 27] # A/B
else:
    bcmPins = [4, 5, 6, 12, 13, 16, 17, 18, 22, 23, 24, 25, 26, 27] # B+

gpioInterface = None

# initial interrupt callback routine that is called when an interrupt pin goes low
def interruptCallback(pin):
    debug('debugGPIO', "interruptCallback", "pin:", pin)
    try:
        sensor = gpioInterface.sensorAddrs[pin]
        state = gpio.input(pin)
        if sensor.interrupt:
            debug('debugGPIO', gpioInterface.name, "calling", sensor.name, "interrupt", state)
            sensor.interrupt(sensor, state)
        debug('debugGPIO', gpioInterface.name, "notifying", sensor.name, state)
        sensor.notify(state)
    except KeyError:
        debug('debugGPIO', gpioInterface.name, "no sensor for interrupt on pin", pin)

# Interface to direct GPIO
class GPIOInterface(Interface):
    def __init__(self, name, interface=None, event=None, input=[], output=[], invert=False):
        Interface.__init__(self, name, interface=interface, event=event)
        global gpioInterface
        gpioInterface = self
        self.input = input
        self.output = output
        self.invert = invert
        # initialize everything
        gpio.setwarnings(False)
        gpio.setmode(gpio.BCM)
        # set I/O direction of pins
        for pin in self.input:
            if pin in bcmPins:
                debug('debugGPIO', self.name, "setup", pin, gpio.IN)
                gpio.setup(pin, gpio.IN, pull_up_down=gpio.PUD_UP)
                gpio.add_event_detect(pin, gpio.FALLING, callback=interruptCallback)
            else:
                debug('debugGPIO', self.name, "ignoring", pin)
        for pin in self.output:               # output pin
            if pin in bcmPins:
                debug('debugGPIO', self.name, "setup", pin, gpio.OUT)
                gpio.setup(pin, gpio.OUT)
                debug('debugGPIO', self.name, "write", pin, 0)
                if self.invert:
                    gpio.output(pin, 1)
                else:
                    gpio.output(pin, 0)
            else:
                debug('debugGPIO', self.name, "ignoring", pin)

    def read(self, addr):
        value = gpio.input(addr)
        if self.invert:
            value = 1 - value
        debug('debugGPIO', self.name, "read", "addr:", addr, "value:", value)
        return value

    def write(self, addr, value):
        debug('debugGPIO', self.name, "write", "addr:", addr, "value:", value)
        if self.invert:
            value = 1 - value
        gpio.output(addr, value)




weatherSampleRate = 10

import time
import threading
import math
import bme680
from ha import *

# BME-680 temperature, humidity, pressure, VOC sensor

class BME680Interface(Interface):
    def __init__(self, name, interface=None, addr=0x77, event=None):
        Interface.__init__(self, name, interface, event=event)
        self.addr = addr
        self.lock = threading.Lock()
        self.correction = .68    # pressure correction - rough approximation in inches
        self.sensor = bme680.BME680(self.addr)
        self.sensor.set_humidity_oversample(bme680.OS_2X)
        self.sensor.set_pressure_oversample(bme680.OS_4X)
        self.sensor.set_temperature_oversample(bme680.OS_8X)
        self.sensor.set_filter(bme680.FILTER_SIZE_3)
        # self.sensor.set_gas_status(bme680.ENABLE_GAS_MEAS)
        # self.sensor.set_gas_heater_temperature(320)
        # self.sensor.set_gas_heater_duration(150)
        # self.sensor.select_gas_heater_profile(0)
        def sample():
            while True:
                with self.lock:
                    try:
                        while not self.sensor.get_sensor_data():
                            time.sleep(.1)
                    except IOError as ex:
                        log(self.name, type(ex).__name__, str(ex))
                time.sleep(weatherSampleRate)
        sampleThread = LogThread(name="sampleThread", target=sample)
        sampleThread.start()

    def read(self, addr):
        with self.lock:
            if addr == "temp":
                value = self.sensor.data.temperature * 9/5 + 32
            elif addr == "humidity":
                value = self.sensor.data.humidity
            elif addr == "dewpoint":
                # https://en.wikipedia.org/wiki/Dew_point
                b = 17.62
                c = 243.12
                g = math.log(self.sensor.data.humidity / 100) + ((b * self.sensor.data.temperature) / (c + self.sensor.data.temperature))
                value = (c * g) / (b - g) * 9/5 + 32
            elif addr == "barometer":
                value = self.sensor.data.pressure  * 0.029529983071445 + self.correction
            elif addr == "voc":
                if self.sensor.data.heat_stable:
                    value = self.sensor.data.gas_resistance
                else:
                    value = 0
            else:
                value = 0
        debug('debugBME680', self.name, addr, value)
        return round(value, 2)



import time
from ha import *

# BMP085 pressure and temp sensor
# Code lifted from Adafruit_BMP085

class BMP085Interface(Interface):

    # Operating Modes
    __BMP085_ULTRALOWPOWER     = 0
    __BMP085_STANDARD          = 1
    __BMP085_HIGHRES           = 2
    __BMP085_ULTRAHIGHRES      = 3

    # BMP085 Registers
    __BMP085_CAL_AC1           = 0xAA  # R   Calibration data (16 bits)
    __BMP085_CAL_AC2           = 0xAC  # R   Calibration data (16 bits)
    __BMP085_CAL_AC3           = 0xAE  # R   Calibration data (16 bits)
    __BMP085_CAL_AC4           = 0xB0  # R   Calibration data (16 bits)
    __BMP085_CAL_AC5           = 0xB2  # R   Calibration data (16 bits)
    __BMP085_CAL_AC6           = 0xB4  # R   Calibration data (16 bits)
    __BMP085_CAL_B1            = 0xB6  # R   Calibration data (16 bits)
    __BMP085_CAL_B2            = 0xB8  # R   Calibration data (16 bits)
    __BMP085_CAL_MB            = 0xBA  # R   Calibration data (16 bits)
    __BMP085_CAL_MC            = 0xBC  # R   Calibration data (16 bits)
    __BMP085_CAL_MD            = 0xBE  # R   Calibration data (16 bits)
    __BMP085_CONTROL           = 0xF4
    __BMP085_TEMPDATA          = 0xF6
    __BMP085_PRESSUREDATA      = 0xF6
    __BMP085_READTEMPCMD       = 0x2E
    __BMP085_READPRESSURECMD   = 0x34

    # Private Fields
    _cal_AC1 = 0
    _cal_AC2 = 0
    _cal_AC3 = 0
    _cal_AC4 = 0
    _cal_AC5 = 0
    _cal_AC6 = 0
    _cal_B1 = 0
    _cal_B2 = 0
    _cal_MB = 0
    _cal_MC = 0
    _cal_MD = 0

    def __init__(self, name, interface, addr=0x77, mode=1, height=0):
        Interface.__init__(self, name, interface)
        correctionAlt = (elevation + height) * 0.3048  # correction height in meters
        self.correction = .68    # fuck it - rough approximation of correction in inches
        self.addr = addr
        # Make sure the specified mode is in the appropriate range
        if ((mode < 0) | (mode > 3)):
          if (self.debug):
            print("Invalid Mode: Using STANDARD by default")
          self.mode = self.__BMP085_STANDARD
        else:
          self.mode = mode
        # Read the calibration data
        self.readCalibrationData()

    def read(self, addr):
        debug('debugTemp', self.name, "read", addr)
        try:
            if addr == "temp":
                return self.readTemperature() * 9 / 5 + 32
            elif addr == "barometer":
                return ((self.readPressure() / 100.0) * 0.0295301) + self.correction
            else:
                return 0
        except:
            return 0

    def readCalibrationData(self):
        # Reads the calibration data from the IC"
        self._cal_AC1 = self.readS16(self.__BMP085_CAL_AC1)   # INT16
        self._cal_AC2 = self.readS16(self.__BMP085_CAL_AC2)   # INT16
        self._cal_AC3 = self.readS16(self.__BMP085_CAL_AC3)   # INT16
        self._cal_AC4 = self.readU16(self.__BMP085_CAL_AC4)   # UINT16
        self._cal_AC5 = self.readU16(self.__BMP085_CAL_AC5)   # UINT16
        self._cal_AC6 = self.readU16(self.__BMP085_CAL_AC6)   # UINT16
        self._cal_B1 = self.readS16(self.__BMP085_CAL_B1)     # INT16
        self._cal_B2 = self.readS16(self.__BMP085_CAL_B2)     # INT16
        self._cal_MB = self.readS16(self.__BMP085_CAL_MB)     # INT16
        self._cal_MC = self.readS16(self.__BMP085_CAL_MC)     # INT16
        self._cal_MD = self.readS16(self.__BMP085_CAL_MD)     # INT16
        self.showCalibrationData()

    def showCalibrationData(self):
        debug("debugBMP085", self.name, "calibration", "AC1 = %6d" % (self._cal_AC1),
                                                       "AC2 = %6d" % (self._cal_AC2),
                                                       "AC3 = %6d" % (self._cal_AC3),
                                                       "AC4 = %6d" % (self._cal_AC4),
                                                       "AC5 = %6d" % (self._cal_AC5),
                                                       "AC6 = %6d" % (self._cal_AC6),
                                                       "B1  = %6d" % (self._cal_B1),
                                                       "B2  = %6d" % (self._cal_B2),
                                                       "MB  = %6d" % (self._cal_MB),
                                                       "MC  = %6d" % (self._cal_MC),
                                                       "MD  = %6d" % (self._cal_MD))

    def readRawTemp(self):
        # Reads the raw (uncompensated) temperature from the sensor
        self.write8(self.__BMP085_CONTROL, self.__BMP085_READTEMPCMD)
        time.sleep(0.005)  # Wait 5ms
        raw = self.readU16(self.__BMP085_TEMPDATA)
        debug("debugBMP085", self.name, "Raw Temp: 0x%04X (%d)" % (raw & 0xFFFF, raw))
        return raw

    def readRawPressure(self):
        # Reads the raw (uncompensated) pressure level from the sensor
        self.write8(self.__BMP085_CONTROL, self.__BMP085_READPRESSURECMD + (self.mode * 2**6))
        if (self.mode == self.__BMP085_ULTRALOWPOWER):
            time.sleep(0.005)
        elif (self.mode == self.__BMP085_HIGHRES):
            time.sleep(0.014)
        elif (self.mode == self.__BMP085_ULTRAHIGHRES):
            time.sleep(0.026)
        else:
            time.sleep(0.008)
        msb = self.readU8(self.__BMP085_PRESSUREDATA)
        lsb = self.readU8(self.__BMP085_PRESSUREDATA+1)
        xlsb = self.readU8(self.__BMP085_PRESSUREDATA+2)
        raw = ((msb * 2**16) + (lsb * 2**8) + xlsb) / 2**(8 - self.mode)
        debug("debugBMP085", self.name, "Raw Pressure: 0x%04X (%d)" % (int(raw) & 0xFFFF, raw))
        return raw

    def readTemperature(self):
        # Gets the compensated temperature in degrees celsius

        # Read raw temp before aligning it with the calibration values
        UT = self.readRawTemp()
        X1 = ((UT - self._cal_AC6) * self._cal_AC5) / 2**15
        X2 = (self._cal_MC * 2**11) / (X1 + self._cal_MD)
        B5 = X1 + X2
        temp = ((B5 + 8) / 2**4) / 10.0
        debug("debugBMP085", self.name, "Calibrated temperature = %f C" % temp)
        return temp

    def readPressure(self):
        # Gets the compensated pressure in pascal
        UT = self.readRawTemp()
        UP = self.readRawPressure()
        self.showCalibrationData()

        # True Temperature Calculations
        X1 = ((UT - self._cal_AC6) * self._cal_AC5) / 2**15
        X2 = (self._cal_MC * 2**11) / (X1 + self._cal_MD)
        B5 = X1 + X2
        debug("debugBMP085", self.name, "X1 = %d" % (X1), "X2 = %d" % (X2), "B5 = %d" % (B5), "True Temperature = %.2f C" % (((B5 + 8) / 2**4) / 10.0))

        # Pressure Calculations
        B6 = B5 - 4000
        X1 = (self._cal_B2 * (B6 * B6) / 2**12) / 2**11
        X2 = (self._cal_AC2 * B6) / 2**11
        X3 = X1 + X2
        B3 = (((self._cal_AC1 * 4 + X3) * 2**self.mode) + 2) / 4
        debug("debugBMP085", self.name, "B6 = %d" % (B6), "X1 = %d" % (X1), "X2 = %d" % (X2), "B3 = %d" % (B3))

        X1 = (self._cal_AC3 * B6) / 2**13
        X2 = (self._cal_B1 * ((B6 * B6) / 2**12)) / 2**16
        X3 = ((X1 + X2) + 2) / 2**2
        B4 = (self._cal_AC4 * (X3 + 32768)) / 2**15
        B7 = (UP - B3) * (50000 / 2**self.mode)
        debug("debugBMP085", self.name, "X1 = %d" % (X1), "X2 = %d" % (X2), "B4 = %d" % (B4), "B7 = %d" % (B7))

        if (B7 < 0x80000000):
          p = (B7 * 2) / B4
        else:
          p = (B7 / B4) * 2

        X1 = (p / 2**8) * (p / 2**8)
        X1 = (X1 * 3038) / 2**16
        X2 = (-7375 * p) / 2**16
        debug("debugBMP085", self.name, "p  = %d" % (p), "X1 = %d" % (X1), "X2 = %d" % (X2))

        p = p + ((X1 + X2 + 3791) / 2**4)
        debug("debugBMP085", self.name, "Pressure = %d Pa" % (p))

        return p

    def readAltitude(self, seaLevelPressure=101325):
        # Calculates the altitude in meters
        altitude = 0.0
        pressure = float(self.readPressure())
        altitude = 44330.0 * (1.0 - pow(pressure / seaLevelPressure, 0.1903))
        debug("debugBMP085", self.name, "Altitude = %d" % (altitude))
        return altitude

    def readU8(self, reg):
        # Read an unsigned byte from the I2C device
        result = self.interface.read((self.addr, reg))
        return result

    def readS8(self, reg):
        # Reads a signed byte from the I2C device
        result = self.interface.read((self.addr, reg))
        if (result > 127):
            return result - 256
        else:
            return result

    def readU16(self, reg):
        # Reads an unsigned 16-bit value from the I2C device
        hibyte = self.interface.read((self.addr, reg))
        result = (hibyte * 2**8) + self.interface.read((self.addr, reg+1))
        return result

    def readS16(self, reg):
        # Reads a signed 16-bit value from the I2C device
        hibyte = self.interface.read((self.addr, reg))
        if (hibyte > 127):
            hibyte -= 256
        result = (hibyte * 2**8) + self.interface.read((self.addr, reg+1))
        return result

    def write8(self, reg, value):
        # Writes an 8-bit value to the specified register/address
        self.interface.write((self.addr, reg), value)



from ha import *

class PowerInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)

    def read(self, theAddr):
        if theAddr.getState():
            return powerTbl[theAddr.name]
        else:
            return 0 





from ha import *

class LoadInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
#        self.sql = """select loads.%s from loads,
#	                    (select max(time) time from loads where date = curdate()) max
#	                    where loads.date = curdate() and loads.time = max.time;	
#                        """
        self.volts = {"Lights":120,
	                  "Plugs":120,
	                  "Appl1":120,
	                  "Cooking":240,
	                  "Appl2":120,
	                  "Ac":240,
	                  "Pool":240,
	                  "Back":240,
	                  "CurrentLoad":240,
	                  "DailyLoad":240,
	                   }

    def read(self, addr):
        try:
#            table = theAddr[0]
#            sensor = theAddr[2].lower()
#            sql = self.sql % (sensor)
#            return self.interface.read(sql) * self.volts[sensor]
            return self.interface.read(addr) * self.volts[addr]
        except:
            return "-"




# Core class definitions

import time
import threading
import copy
import sys
import json
from collections import OrderedDict
from .config import *
from .environment import *
from .logging import *
from .debugging import *

# states
off = 0
Off = 0
on = 1
On = 1

### general utility functions ###

# get the value of a variable from a file
def getValue(fileName, item):
    return json.load(open(fileName))[item]

# turn an item into a list if it is not already
def listize(x):
    return x if isinstance(x, list) else [x]

# normalize state values from boolean to integers
def normalState(value):
    if value == True: return On
    elif value == False: return Off
    else: return value

# Compare two state dictionaries and return a dictionary containing the items
# whose values don't match or aren't in the old dict.
# If an item is in the old but not in the new, optionally include the item with value None.
def diffStates(old, new, deleted=True):
    diff = copy.copy(new)
    for key in list(old.keys()):
        try:
            if new[key] == old[key]:
                del diff[key]   # values match
        except KeyError:        # item is missing from the new dict
            if deleted:         # include deleted item in output
                diff[key] = None
    return diff

# create a Resource from a serialized dict
def loadResource(classDict, globalDict):
    def parseClass(classDict):
        args = classDict["args"]
        argStr = ""
        for arg in list(args.keys()):
            argStr += arg+"="
            if isinstance(args[arg], dict):     # argument is a class
                argStr += parseClass(args[arg])+", "
            elif isinstance(args[arg], str):    # arg is a string
                argStr += "'"+args[arg]+"', "
            elif not arg:                       # arg is None
                argStr += "None"
            else:                               # arg is numeric or other
                argStr += str(args[arg])+", "
        return classDict["class"]+"("+argStr[:-2]+")"
    localDict = {}
    exec("resource = "+parseClass(classDict), globalDict, localDict)
    return localDict["resource"]

# Base class for everything
class Object(object):
    def __init__(self):
        self.className = self.__class__.__name__    # Used to optionally override the real class name in dump()

    # dump the resource attributes to a serialized dict
    def dump(self, expand=False):
        return {"class": self.className,
                "args": self.dict(expand)}

# Base class for Resources
class Resource(Object):
    def __init__(self, name, type):
        Object.__init__(self)
        try:
            if self.name:   # init has already been called for this object - FIXME
                return
        except AttributeError:
            self.name = name
            self.type = type
            self.enabled = True
            self.collections = {}   # list of collections that include this resource
            debug('debugObject', self.__class__.__name__, self.name, "created")

    def enable(self):
        self.enabled = True

    def disable(self):
        if self.enabled:
            self.enabled = False
            # nullify cached states
            for collection in list(self.collections.values()):
                collection.states[self.name] = None
            if self.interface:
                self.interface.states[self.name] = None

    # add this resource to the specified collection
    def addCollection(self, collection):
        self.collections[collection.name] = collection

    # remove this resource from the specified collection
    def delCollection(self, collection):
        del self.collections[collection.name]

    # jquery doesn't like periods in names
    def jqName(self):
        return self.name.replace(".", "_")

    def __str__(self):
        return self.name

# Base class for Interfaces
class Interface(Resource):
    def __init__(self, name, interface=None, type="interface", event=None):
        Resource.__init__(self, name, type)
        self.interface = interface
        # sensor state change event
        if event != None:                   # use the specified one
            self.event = event
        elif self.interface:
            self.event = interface.event    # otherwise inherit event from this interface's interface
        else:
            self.event = None
        self.sensors = {}       # sensors using this instance of the interface by name
        self.sensorAddrs = {}   # sensors using this instance of the interface by addr
        self.states = {}        # sensor state cache

    def start(self):
        return True

    def stop(self):
        return True

    def read(self, addr):
        return None

    def write(self, addr, value):
        return True

    def dump(self):
        return None

    # add a sensor to this interface
    def addSensor(self, sensor):
        debug('debugObject', self.__class__.__name__, self.name, "addSensor", sensor.name)
        self.sensors[sensor.name] = sensor
        self.sensorAddrs[sensor.addr] = sensor
        self.states[sensor.addr] = None
        sensor.event = self.event

    # Trigger the sending of a state change notification
    def notify(self):
        if self.event:
            self.event.set()

# Resource collection
class Collection(Resource, OrderedDict):
    def __init__(self, name, resources=[], aliases={}, type="collection", event=None, start=False):
        Resource.__init__(self, name, type)
        OrderedDict.__init__(self)
        self.lock = threading.Lock()
        self.states = {}    # cache of current sensor states
        for resource in resources:
            self.addRes(resource)
        self.aliases = aliases
        if event:
            self.event = event
        else:
            self.event = threading.Event()
        debug('debugCollection', self.name, "aliases:", self.aliases)
        if start:
            self.start()

    def start(self):
        # thread to periodically poll the state of the resources in the collection
        def pollStates():
            debug('debugCollection', self.name, "starting resource polling")
            resourcePollCounts = {}
            while True:
                stateChanged = False
                with self.lock:
                    for resource in list(self.values()):
                        try:
                            if not resource.event:    # don't poll resources with events
                                if resource.type not in ["schedule", "collection", "task"]:   # skip resources that don't have a state
                                    if resource.name not in list(resourcePollCounts.keys()):
                                        resourcePollCounts[resource.name] = resource.poll
                                        self.states[resource.name] = resource.getState()
                                        debug('debugCollectionState', self.name, resource.name, resource.poll, resource.getState())
                                    if resourcePollCounts[resource.name] == 0:          # count has decremented to zero
                                        resourceState = resource.getState()
                                        debug('debugCollectionState', self.name, resource.name, self.states[resource.name], resourceState)
                                        if resourceState != self.states[resource.name]: # save the state if it has changed
                                            self.states[resource.name] = resourceState
                                            stateChanged = True
                                        resourcePollCounts[resource.name] = resource.poll
                                    else:   # decrement the count
                                        resourcePollCounts[resource.name] -= 1
                            else:
                                debug('debugCollectionState', self.name, "skipping", resource.name)
                        except Exception as ex:
                            logException(self.name+" pollStates", ex)
                if stateChanged:    # at least one resource state changed
                    debug('debugCollectionState', self.name, "state changed")
                    self.event.set()
                    stateChanged = False
                time.sleep(1)
            debug('debugCollection', self.name, "ending resource polling")
        # initialize the resource state cache
        for resource in list(self.values()):
            if resource.type not in ["schedule", "collection"]:   # skip resources that don't have a state
                try:
                    self.states[resource.name] = resource.getState()    # load the initial state
                except Exception as ex:
                    logException(self.name+" start", ex)
        pollStatesThread = LogThread(name="pollStatesThread", target=pollStates)
        pollStatesThread.start()

    # Add a list of resources to this collection
    def addRes(self, resources, state=None):
        if not isinstance(resources, list):
            resources = [resources]
        for resource in resources:
            with self.lock:
                try:
                    debug('debugCollection', self.name, "adding", resource.name)
                    self.__setitem__(str(resource), resource)
                    resource.addCollection(self)
                    self.states[resource.name] = state
                except Exception as ex:
                    logException(self.name+" addRes", ex)

    # Delete a list of resources from this collection
    def delRes(self, names):
        if not isinstance(names, list):
            names = [names]
        for name in names:
            with self.lock:
                try:
                    debug('debugCollection', self.name, "deleting", name)
                    del self.states[name]
                    self.__getitem__(name).delCollection(self)
                    self.__delitem__(name)
                except Exception as ex:
                    logException(self.name+" delRes", ex)

    # Get a resource from the collection
    # Return dummy sensor if not found
    def getRes(self, name, dummy=True):
        try:
            return self.__getitem__(name)
        except KeyError:
            if dummy:
                return Sensor(name)
            else:
                raise

    # Return the list of resources that have the names specified in the list
    def getResList(self, names):
        resList = []
        for name in names:
            try:
                resList.append(self.getRes(name))
            except:
                pass
        return resList

    # Return a list of resource references that are members of the specified group
    # in order of addition to the table
    def getGroup(self, group):
        debug('debugCollection', self.name, "getGroup", group)
        resourceList = []
        for resourceName in list(self.keys()):
            resource = self.__getitem__(resourceName)
            if group in listize(resource.group):
                resourceList.append(resource)
        return resourceList

    # get the current state of all sensors in the resource collection
    def getStates(self, wait=False):
        debug('debugCollection', self.name, "getStates", wait)
        if self.event and wait:
            self.event.clear()
            self.event.wait()
        return copy.copy(self.states)

    # set the state of the specified sensor in the cache
    def setState(self, sensor, state):
        self.states[sensor.name] = state

    # set state values of all sensors into the cache
    def setStates(self, states):
        debug('debugCollection', self.name, "setStates", states)
        for sensor in list(states.keys()):
            self.states[sensor] = states[sensor]

    # Trigger the sending of a state change notification
    def notify(self):
        if self.event:
            self.event.set()

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        return {"name":self.name,
                "type": self.type,
                "resources":([attr.dump(expand) for attr in list(self.values())] if expand else list(self.keys()))}

# A Sensor represents a device that has a state that is represented by a scalar value.
# The state is associated with a unique address on an interface.
# Sensors can also optionally be associated with a group and a physical location.
class Sensor(Resource):
    def __init__(self, name, interface=None, addr=None, type="sensor",
                 factor=1, offset=0, resolution=0,
                 poll=10, event=None, persistence=None, interrupt=None,
                 location=None, group="", label=None):
        try:
            if self.type:   # init has already been called for this object - FIXME
                return
        except AttributeError:
            Resource.__init__(self, name, type)
            self.interface = interface
            self.addr = addr
            if self.interface:
                self.interface.addSensor(self)
            self.resolution = resolution
            self.factor = factor
            self.offset = offset
            self.poll = poll
            if event:
                self.event = event
            elif self.interface:       # inherit the event from the interface if not specified
                self.event = self.interface.event
            else:
                self.event = None
            self.persistence = persistence
            self.interrupt = interrupt
            self.location = location
            self.group = listize(group)
            if label:
                self.label = label
            else:
                self.label = self.name.capitalize()
            self.__dict__["state"] = None   # dummy class variable so hasattr() returns True
            # FIXME - use @property

    # Return the state of the sensor by reading the value from the address on the interface.
    def getState(self, missing=None):
        debug('debugSensor', self.name, "getState")
        state = (normalState(self.interface.read(self.addr)) if self.interface else None)
        try:
            return round(state * self.factor + self.offset, self.resolution)
        except TypeError:
            return state

    # Trigger the sending of a state change notification
    def notify(self, state=None):
        if not state:
            state = self.getState()
        for collection in list(self.collections.keys()):
            self.collections[collection].setState(self, state)
        if self.event:
            self.event.set()

    # Define this function for sensors even though it does nothing
    def setState(self, state, wait=False):
        return False

    # override to handle special cases of state
    def __getattribute__(self, attr):
        if attr == "state":
            return self.getState()
        else:
            return Resource.__getattribute__(self, attr)

    # override to handle special case of state
    def __setattr__(self, attr, value):
        if attr == "state":
            self.setState(value)
        else:
            Resource.__setattr__(self, attr, value)

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        return {"name":self.name,
                "interface":(self.interface.name if self.interface else None),
                "addr":self.addr,
                "type":self.type,
                "resolution": self.resolution,
                "poll": self.poll,
                "persistence": str(self.persistence),
                "location":self.location,
                "group":self.group,
                "label":self.label}

# A Control is a Sensor whose state can be set
class Control(Sensor):
    def __init__(self, name, interface=None, addr=None, type="control", stateSet=None, **kwargs):
        Sensor.__init__(self, name, interface=interface, addr=addr, type=type, **kwargs)
        self.stateSet = stateSet  # optional callback when state is set

    # Set the state of the control by writing the value to the address on the interface.
    def setState(self, state, wait=False):
        debug('debugState', "Control", self.name, "setState ", state)
        if self.enabled:
            self.interface.write(self.addr, state)
            self.notify(state)
            if self.stateSet:
                self.stateSet(self, state)
            return True
        else:
            return False



poolValveTravelTime = 35

import time
import threading
from ha.interfaces.gpioInterface import *
from ha import *

# valve states
valveMoving = 4

class ValveInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.travelTime = poolValveTravelTime
        self.timers = {}
        self.lock = threading.Lock()

    def read(self, addr):
        try:
            return self.states[addr]
        except:
            return 0

    def write(self, addr, value):
#        self.newValue = value
        debug('debugValves', self.name, "start", addr, self.states[addr])
        self.states[addr] = valveMoving
        self.sensorAddrs[addr].notify()
        # cancel the timer if it is running
        if self.timers[addr]:
            self.timers[addr].cancel()
        with self.lock:
            # start the motion
            debug('debugValves', self.name, "motion", addr, self.states[addr])
            self.interface.write(addr, value)
        # clean up and set the final state when motion is finished
        def doneMoving():
            with self.lock:
                self.states[addr] = value # done moving
                self.sensorAddrs[addr].notify()
                debug('debugValves', self.name, "done", addr, self.states[addr])
        self.timers[addr] = threading.Timer(self.travelTime, doneMoving)
        self.timers[addr].start()

    def addSensor(self, sensor):
        Interface.addSensor(self, sensor)
        self.timers[sensor.addr] = None
        self.states[sensor.addr] = 0    # initialize state to 0



from ha import *
import threading

# analog temp sensor

class AnalogTempInterface(Interface):
    def __init__(self, name, interface):
        Interface.__init__(self, name, interface)
        self.lock = threading.Lock()

    def read(self, addr):
        debug('debugAnalogTemp', self.name, "read", addr)
        try:
            with self.lock:
                mVolts = self.interface.read(addr)
            volts = float(mVolts)/1000 + .0005
            tempC = 28.25 * (2.52 - volts)
            tempF = tempC * 9 / 5 + 32
            debug('debugAnalogTemp', self.name, "volts", volts, "tempC", tempC, "tempF", tempF)
            return tempF
        except:
            return 0




# Calculate rainfall using the Peet Tipping Bucket rain gauge
rainIncr = .01

import time
import threading
from ha import *
from ha.interfaces.gpioInterface import *

# global variables
rainSamples = []        # list of times of rainIncr inches sampled
sampleLock = threading.Lock()
sampleEvent = threading.Event()

# rain gauge contact closure interrupt routine
def rainInterrupt(sensor, state):
    global rainSamples, sampleLock, sampleEvent
    curTime = time.time()
    if state == 0:
        with sampleLock:
            rainSamples.append(curTime)
        sampleEvent.set()

class RainInterface(Interface):
    def __init__(self, name, interface, rainGauge, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.rainGauge = rainGauge    # rain gauge contact closure sensor
        self.rainGauge.interrupt = rainInterrupt
        debug("debugRain", self.name, "rainGauge:", self.rainGauge.name)

    def start(self):
        global rainSamples, sampleLock, sampleEvent
        debug("debugRain", self.name, "start")
        # read persistent data
        try:
            samples = self.interface.read("rainSamples")
            if not samples:
                samples = []
        except:
            samples = []
        with sampleLock:
            rainSamples = samples
        # thread to save persistent data
        sampleEvent.clear()
        def saveSamples():
            while True:
                sampleEvent.wait()
                with sampleLock:
                    samples = rainSamples
                debug("debugRain", self.name, "samples", str(samples))
                self.interface.write("rainSamples", samples)
                sampleEvent.clear()
        saveSamplesThread = LogThread(name="saveSamplesThread", target=saveSamples)
        saveSamplesThread.start()

    def read(self, addr):
        global rainSamples, sampleLock, sampleEvent
        # determine how many samples to count
        now = time.time()
        if addr == "minute":
            cutoff = now - 60               # sample period is the last minute
        elif addr == "hour":
            cutoff = now - 3600             # sample period is the last hour
        elif addr == "today":
            today = time.localtime(now)     # sample period is since midnight
            cutoff = time.mktime((today.tm_year,today.tm_mon,today.tm_mday,0,0,0,0,0,0))
        else:
            cutoff = time.time()
        # debug("debugRain", self.name, "read", "addr:", addr, "now:", now, "cutoff:", cutoff)
        # copy the sample list
        with sampleLock:
            samples = rainSamples
        rainTotal = 0.0
        for sample in reversed(samples):
            if sample > cutoff: # sample is within the sample period
                rainTotal += rainIncr
            else:               # return the total
                break
        # debug("debugRain", self.name, "read", "rainTotal:", rainTotal)
        return rainTotal

    def write(self, addr, value):
        global rainSamples, sampleLock, sampleEvent
        debug("debugRain", self.name, "write", "addr:", addr, "value:", value)
        if addr == "reset":
            # reset the sample list
            with sampleLock:
                rainSamples = []
            sampleEvent.set()



flickerFreq = 100   # frequency of flickering per second

import random
import time
import threading
from ha import *

class LedInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.running = False
        self.states = {}

    def read(self, addr):
        debug('debugLed', self.name, "read", addr)
        return self.states[addr]

    def write(self, addr, value):
        debug('debugLed', self.name, "write", addr, value)
        self.states[addr] = value
        def flickerThread():
            debug('debugLed', self.name, "starting", addr, value)
            while self.running:
               self.interface.write(addr, 1)
               time.sleep(random.random()/flickerFreq) 
               self.interface.write(addr, 0)
               time.sleep(random.random()/flickerFreq) 
            debug('debugLed', self.name, "ending", addr, value)
        if (value == 0) or (value == 1):
            if self.running:
                self.running = False
                time.sleep(2./flickerFreq)  # wait for thread to finish
            self.interface.write(addr, value)
        elif value == 2:
            self.thread = LogThread(name="self.thread", target=flickerThread)
            self.running = True
            self.thread.start()    





# Logging functions

import syslog
import os
import time
import threading
import traceback
from .environment import *

# standard timestamp
def timeStamp(fmt):
    return time.strftime(fmt, time.localtime())

# log a message to syslog or stdout
def log(*args):
    message = args[0]+" "   # first argument is the object doing the logging
    for arg in args[1:]:
        message += arg.__str__()+" "
    if sysLogging:
        syslog.syslog(message)
    else:
        print(timeStamp("%b %d %H:%M:%S")+" "+message)

# log the traceback for an exception
def logException(name, ex):
    tracebackLines = traceback.format_exception(None, ex, ex.__traceback__)
    log(name+":")
    for tracebackLine in tracebackLines:
        log(tracebackLine)

# thread object that logs a stack trace if there is an uncaught exception
class LogThread(threading.Thread):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.runTarget = self.run
        self.run = self.logThread

    def logThread(self):
        try:
            self.runTarget()
        except Exception as ex:
            logException("thread "+threading.currentThread().name, ex)

# convenience function to create and start a thread
def startThread(name, target, **kwargs):
    thread = LogThread(name=name, target=target, **kwargs)
    thread.start()



from ha import *

# One Wire File System interface

class OWFSInterface(Interface):
    def __init__(self, name, interface=None, home="/mnt/1wire/", event=None):
        Interface.__init__(self, name, interface, event=event)
        self.home = home

    def read(self, addr):
        debug('debugTemp', self.name, "read", addr)
        try:
            with open(self.home+addr+"/temperature") as owfs:
                value = owfs.read()
                try:
                    return float(value)
                except:
                    return value
        except:
            return 0




# Read configuration files and set global variables
localController = "house"

import os
import socket
import time
from .environment import *
from .logging import *

global debugEnable
debugEnable = False
debugConf = False

try:
    for configFileName in ['ha.conf']: #os.listdir(configDir):
        try:
            with open(configDir+configFileName) as configFile:
                configLines = [configLine.rstrip('\n') for configLine in configFile]
            debugConfOpen = True
            for configLine in configLines:
                if (len(configLine) > 0) and (configLine[0] != "#"):
                    try:
                        (key, value) = configLine.split("=")
                        exec("global "+key, globals())
                        exec(configLine, globals())
                        if debugEnable and debugConf:   # can't use debugging module because of circular references
                            if debugConfOpen:   # log the file open retroactively if debugConf got set
                                log("config open", configFileName)
                                debugConfOpen = False
                            log("config read", configLine)
                    except:
                        log("config", "error evaluating", configLine)
        except:
            log("config", "error reading", configDir+configFileName)
except:
    log("config", "no config directory", configDir)

def waitForDns():
    wasWaiting = False
    while True:
        try:
            metricsHost = socket.gethostbyname(localController)
            if wasWaiting:
                log("DNS is up")
            return
        except:
            log("Waiting for DNS")
            wasWaiting = True
            time.sleep(1)



from .config import *
from .logging import *

# log a debug message conditioned on a specified global variable
def debug(*args):
    try:
        if debugEnable:   # global debug flag enables debugging
            if globals()[args[0]]:  # only log if the specified debug variable is True
                log(*args[1:])
    except KeyError:
        pass

# log a stack trace conditioned on a specified global variable
def debugTraceback(debugType, debugName):
    try:
        if debugEnable:   # global debug flag enables debugging
            if globals()[debugType]:
                s = inspect.stack()
                for f in s:
                    log(debugName, f[1], f[2], f[3], f[4])
    except KeyError:
        pass




import time
from math import log as ln
from ha import *

# HIH-6130/6131 humidity sensor

class HIH6130Interface(Interface):
    def __init__(self, name, interface, addr=0x27):
        Interface.__init__(self, name, interface)
        self.addr = addr

    def read(self, addr):
        debug('debugHIH6130', self.name, "read", addr)
        try:
            data = self.interface.readBlock((self.addr, 0), 4)
            status = (data[0] & 0xc0) >> 6
            humidity = ((((data[0] & 0x3f) << 8) + data[1]) * 100.0) / 16383.0
            temp = (((data[2] & 0xff) << 8) + (data[3] & 0xfc)) / 4
            tempC = (temp / 16384.0) * 165.0 - 40.0
            tempF = tempC * 1.8 + 32
            # https://en.wikipedia.org/wiki/Dew_point
            # 0C <= T <= +50C
            b = 17.368
            c = 238.88
            gamma = ln(humidity / 100) + ((b * tempC) / (c + tempC))
            dewpointC = c * (gamma) / (b - gamma)
            dewpointF = dewpointC * 1.8 + 32
            debug("debugHIH6130", "humidity:", humidity, "tempC:", tempC, "tempF:", tempF, "dewpointC:", dewpointC, "dewpointF:", dewpointF)
            if addr == "humidity":
                return humidity
            elif addr == "temp":
                return tempF
            elif addr == "dewpoint":
                return dewpointF
            else:
                return 0
        except:
            return 0




import minimalmodbus
import time
import threading
from ha import *

minimalmodbus.BAUDRATE = 9600
minimalmodbus.TIMEOUT = 0.5

class ModbusInterface(Interface):
    def __init__(self, name, interface=None, event=None, device="", addr=1):
        Interface.__init__(self, name, interface=interface, event=event)
        self.device = device
        self.addr = addr
        self.instrument = minimalmodbus.Instrument(self.device, self.addr)

    def start(self):
        def readStates():
            debug("debugModbus", "starting readStates")
            while True:
                for sensor in list(self.sensors.values()):
                    debug("debugModbus", "readState", sensor.name)
                    try:
                        self.states[sensor.addr] = self.instrument.read_register(sensor.addr)
                    except Exception as ex:
                        log(self.name, "read exception", sensor.name, sensor.addr, type(ex).__name__, str(ex))
                    time.sleep(1)
                debug("debugModbus", "readStates sleeping")
                time.sleep(60 - len(self.sensors))
        readStateThread = LogThread(name="readStateThread", target=readStates)
        readStateThread.start()

    def read(self, addr):
        return self.states[addr]

    def write(self, addr, value):
        try:
            self.instrument.write_register(sensor.addr, value)
        except Exception as ex:
            log(self.name, "write exception", sensor.name, sensor.addr, type(ex).__name__, str(ex))



# Samsung TV remote interface
# https://github.com/Ape/samsungctl

import samsungctl
from ha import *

keyCmds = {
    "off": "KEY_POWEROFF",
    "sleep": "KEY_POWEROFF",
    "sleeping": "KEY_POWEROFF",

    "mute": "KEY_MUTE",
    "dumb": "KEY_MUTE",
    "stun": "KEY_MUTE",
    "unmute": "KEY_MUTE",
    "talkative": "KEY_MUTE",

    "antenna": "KEY_TV",                        # antenna
    "broadcast": "KEY_TV",                      # antenna

    "roku": "KEY_EXT20",                        # HDMI1

    "chromecast": "KEY_AUTO_ARC_PIP_WIDE",      # HDMI2
    "chrome": "KEY_AUTO_ARC_PIP_WIDE",          # HDMI2

    "iphone": "KEY_AUTO_ARC_PIP_RIGHT_BOTTOM",  # HDMI3

    "hdmi": "KEY_AUTO_ARC_AUTOCOLOR_FAIL",      # HDMI4

    "component": "KEY_COMPONENT1",              # component
    "pc": "KEY_PCMODE",                         # VGA
}

config = {
    "name": "ha",
    "description": "PC",
    "id": "",
    "host": "",
    "port": 55000,
    "method": "legacy",
    "timeout": 0,
}

class SamsungInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)

    def read(self, addr):
        return ""

    def write(self, addr, value):
        debug('debugSamsung', self.name, "write", addr, value)
        try:
            if value == 0:
                value = "off"
            keyCmd = keyCmds[value.lower()]
            config["host"] = addr
            samsungctl.Remote(config).control(keyCmd)
        except KeyError:
            log(self.name, addr, "unknown command", value)




import subprocess
import time
from ha import *

# Interface to various OS parameters
class OSInterface(Interface):
    def __init__(self, name, interface=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)

    def read(self, addr):
        addrParts = addr.split()
        if addrParts[0] == "cpuTemp":
            return float(subprocess.check_output("vcgencmd measure_temp", shell=True)[5:-3])
        elif addrParts[0] == "cpuLoad":
            with open('/proc/stat') as f:
                fields = [float(column) for column in f.readline().strip().split()[1:]]
            last_idle, last_total = fields[3], sum(fields)
            time.sleep(1)
            with open('/proc/stat') as f:
                fields = [float(column) for column in f.readline().strip().split()[1:]]
            idle, total = fields[3], sum(fields)
            idle_delta, total_delta = idle - last_idle, total - last_total
            last_idle, last_total = idle, total
            return 100.0 * (1.0 - idle_delta / total_delta)
        elif addrParts[0] == "diskUse":
            useParts = subprocess.check_output("df "+addrParts[1], shell=True).decode().split("\n")[1].split()
            return 100 * int(useParts[2]) / (int(useParts[2]) + int(useParts[3]))
        elif addrParts[0] == "ssid":
            try:
                return subprocess.check_output("iwconfig "+addrParts[1]+"|grep ESSID", shell=True).decode().strip("\n").split(":")[-1].split("/")[0].strip().strip('"')
            except subprocess.CalledProcessError:
                return ""
        elif addrParts[0] == "ipAddr":
            try:
                return subprocess.check_output("ifconfig "+addrParts[1]+"|grep inet\ ", shell=True).decode().strip("\n").split()[1]
            except subprocess.CalledProcessError:
                return ""
        elif addrParts[0] == "uptime":
            return " ".join(c for c in subprocess.check_output("uptime", shell=True).decode().strip("\n").split(",")[0].split()[2:])
        else:
            return None



import os
import subprocess
import time
from ha import *

class VideoInterface(Interface):
    def __init__(self, name, interface, videoControl, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.videoControl = videoControl
        self.state = 0
        self.pid = 0

    def read(self, addr):
        return self.state

    def write(self, addr, value):
        if value != self.state:
            if value:
                video = self.videoControl.getState()
                cmd = "/usr/bin/mplayer -really-quiet -fs -lavdopts threads=4 -loop 0 -vo fbdev /video/"+video+".mp4"
                self.pid = subprocess.Popen(cmd, shell=True).pid+1
                log("pid", self.pid)
            else:
                if self.pid:
                    pid = subprocess.Popen("kill "+str(self.pid), shell=True)
                time.sleep(1)
                pid = subprocess.Popen("dd if=/dev/zero of=/dev/fb0", shell=True)
                pid = subprocess.Popen("setterm -cursor off > /dev/tty1", shell=True)
            self.state = value



import threading
import time
from ha import *

class TempInterface(Interface):
    def __init__(self, name, interface=None, sample=1):
        Interface.__init__(self, name, interface)
        self.sample = sample

    def start(self):
        # data sampling thread
        def readData():
            debug('debugTemp', self.name, "readData started")
            temp = self.readData()    # initialize states with a reading
            samples = 1
            lastMinute = -1
            # sample once per second and update states with the average every minute
            while True:
                thisMinute = int(time.strftime("%M"))
                debug('debugTemp', self.name, "thisMinute", thisMinute, "lastMinute", lastMinute)
                if thisMinute != lastMinute:                    # start of a new minute
                    for addr in list(self.sensorAddrs.keys()):    # average the readings
                        self.states[addr] = temp[addr]/samples
                        temp[addr] = 0.0
                    if self.event:
                        self.event.set()
                    debug('debugTemp', self.name, "state", self.states, "samples", samples)
                    samples = 0
                    lastMinute = thisMinute
                sample = self.readData()            # take another sample
                for addr in list(sample.keys()):
                    try:
                        temp[addr] += sample[addr]
                    except KeyError:                # a sensor was added during the sample period
                        temp[addr] = sample[addr]
                samples += 1
                time.sleep(self.sample)

        readStatesThread = LogThread(name="readStatesThread", target=readData)
        readStatesThread.start()

    def read(self, addr):
        debug('debugTemp', self.name, "read", addr)
        return self.states[addr]

    # take a reading of all sensors
    def readData(self):
        temp = {}
        for addr in list(self.sensorAddrs.keys()):
            temp[addr] = self.interface.read(addr)
        debug('debugTemp', self.name, "readData", temp)
        return temp
